# -*- coding: utf-8 -*- 
# @Time    : 04/04/2024 
# @Author  : nzooherd 
# @File    : object_store.py 
# @Software: Visual Studio Code

import asyncio
from logging import getLogger
import os
import aioboto3
from botocore.exceptions import ClientError

from nzooherd_pyzen.classes import Singleton
from nzooherd_pyzen.collection import bulk_to_batch

from aspect.task import task_logger
from config import Config

logger = getLogger()

class ObjectStore(metaclass=Singleton):

    def __init__(self) -> None:
        self.s3bucket = "sayabot"
        self.s3client = lambda: aioboto3.Session().client("s3",
            region_name="ap-northeast-1",
            aws_access_key_id="AKIA22QJDJXZVJTUIUXN",
            aws_secret_access_key="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"
        )

        self.r2bucket = "sayabot"
        self.r2client = lambda: aioboto3.Session().client("s3",
            endpoint_url = f"https://{Config.CLOUDFLARE_ACCOUNT_ID}.r2.cloudflarestorage.com",
            aws_access_key_id = Config.CLOUDFLARE_R2_ACCESS_KEY,
            aws_secret_access_key = Config.CLOUDFLARE_R2_SECRET_KEY,
            region_name = "auto"
        )

        self.client = lambda use_s3: self.s3client() if use_s3 else self.r2client()
        self.bucket = lambda use_s3: self.s3bucket if use_s3 else self.r2bucket

        self.part_size_limit = 64 * 1024 * 1024
        self.multipar_upload_limite = self.part_size_limit * 8

    async def head_object(self, key: str, use_s3=False):
        async with self.client(use_s3) as s3:
            try:
                return await s3.head_object(Bucket=self.bucket(use_s3), Key=key)
            except ClientError:
                return None


    @task_logger
    async def upload_object(self, key: str, file_path: str, use_s3=False, **kwargs):
        if os.path.getsize(file_path) > self.multipar_upload_limite:
            await self.upload_multiple_part(key, file_path, use_s3=use_s3, **kwargs)
        else:
            await self.upload_single_part(key, file_path, use_s3=use_s3, **kwargs) 

    async def upload_single_part(self, key: str, file_path: str, use_s3=False, **kwargs):
        """
        """
        with open(file_path, 'rb') as f:
            async with self.client(use_s3) as s3:
                await s3.put_object(Body=f, Bucket=self.bucket(use_s3), Key=key, **kwargs)
                

    async def upload_multiple_part(self, key: str, file_path: str, use_s3=False, **kwargs):
        """
        """
        async with self.client(use_s3) as s3:
            bucket = self.bucket(use_s3)
            with open(file_path, 'rb') as fp:
                init_response = await s3.create_multipart_upload(Bucket=bucket, Key=key, **kwargs)
                upload_id = init_response['UploadId']
                file_size = os.path.getsize(file_path)
                part_size = int(self.part_size_limit)
                part_count = (file_size +  part_size - 1) // part_size 

                parts = []
                for sequence in bulk_to_batch(range(part_count), 10):
                    futures = []
                    part_numbers = []
                    for i in sequence:
                        part_number = i + 1
                        start = part_size * i
                        part_size = min(part_size, file_size - start)
                        body = fp.read(part_size)
                        # 上传分片
                        part_numbers.append(part_number)
                        futures.append(s3.upload_part(Bucket=bucket, Key=key, 
                                                PartNumber=part_number, UploadId=upload_id, Body=body))
                
                    results = await asyncio.gather(*futures)
                    for part_number, item in zip(part_numbers, results):
                        parts.append({'PartNumber': part_number, 'ETag': item['ETag']})

                # 完成分片上传
                await s3.complete_multipart_upload(Bucket=bucket, Key=key,
                                                         UploadId=upload_id, MultipartUpload={'Parts': parts})
                logger.info(f"File uploaded successfully: {key}")

    def permant_url(self, key: str, use_s3=False) -> str:
        """
        Generate permanent download url.

        Because cloudflare R2 only supports presigned url for 7 days. So generate a url to third party web service. On the third party service, 
        by http redirect, redirect user to a not expired url.
        """
        if not use_s3:
            return f"https://saya.zehornod.workers.dev/{key}"
        return f"https://{self.bucket(use_s3)}.s3.ap-northeast-1.amazonaws.com/{key}"
    
